import inspect
import sys

from collections import defaultdict
from math import ceil, floor
from op_service.type_validation import validate
from op_service.type_validation_functions import int_range, non_empty_str, nullable, numeric_range, one_of, regex
from random import Random


def _validate_build_fn(fn):
  if not inspect.iscoroutinefunction(fn):
    raise Exception('build function should be a coroutine')
  if len(inspect.signature(fn).parameters) != 1:
    raise Exception('build function should take exactly one parameter')


def _validate_audio_choice_spec(spec):
  validate(spec, { 'options': [ { 'name': non_empty_str, 'src': nullable(non_empty_str), '[default]': bool } ] })
  if len(spec['options']) < 2:
    raise Exception('must provide at least two options for an AUDIO_CHOICE input')
  if len(tuple(x for x in spec['options'] if not x.get('src'))) > 1:
    raise Exception("only one option can have an empty 'src' for an AUDIO_CHOICE input")
  if len(tuple(x for x in spec['options'] if x.get('default'))) > 1:
    raise Exception('cannot set more than one default for an AUDIO_CHOICE input')

def _validate_img_group_spec(spec):
  validate(spec, { '[min]': int_range('[0,∞)'), 'max': int_range('[2,∞)') })
  if spec.get('min', 0) > spec['max']:
    raise Exception("'min' cannot be greater than 'max' for an IMG_GROUP input")

# https://gist.github.com/olmokramer/82ccce673f86db7cda5e
CSS_COLOR_REGEX = '^(#[0-9a-fA-F]{3}|#(?:[0-9a-fA-F]{2}){2,4}|(rgb|hsl)a?\((-?\d+%?[,\s]+){2,3}\s*[\d\.]+%?\))$'

# TODO: this is redundant - the front-end is the authoritative source of these
FONTS = [
  'Architects Daughter',
  'Average',
  'Balsamiq Sans',
  'Caveat',
  'Chelsea Market',
  'Dancing Script',
  'Fondamento',
  'Indie Flower',
  'Nunito',
  'Rock Salt',
  'Sacramento',
  'Amatic SC',
  'Average',
  'Berkshire Swash',
  'Bubblegum Sans',
  'Chewy',
  'Emilys Candy',
  'Fredericka the Great',
  'Grand Hotel',
  'Homemade Apple',
  'Kaushan Script',
  'Lobster',
  'Luckiest Guy',
  'Permanent Marker',
  'Ranchers'
]

def _validate_color_options(options):
  if isinstance(options, str):
    if options != 'DEFAULT':
      raise Exception("'color_options' must be None, the string 'DEFAULT', or a list of valid CSS color strings")
    return
  validate(options, [ regex(CSS_COLOR_REGEX) ])
  if len(options) < 2:
    raise Exception("must provide at least two 'color_options'")

def _validate_font_options(options):
  if isinstance(options, str):
    if options != 'DEFAULT':
      raise Exception("'font_options' must be None, the string 'DEFAULT', or a list of valid fonts")
    return
  validate(options, [ one_of(*FONTS) ])
  if len(options) < 2:
    raise Exception("must provide at least two 'font_options'")

def _validate_text_spec(spec):
  validate(
    spec,
    {
      'bounds': { 'w': int_range('(0,∞)'), 'h': int_range('(0,∞)') },
      'font_size': int,
      'min_font_size': int,
      'text': non_empty_str,
      'font': one_of(*FONTS),
      'color': regex(CSS_COLOR_REGEX),
      'bold': bool,
      'italic': bool,
      'underline': bool,
      'background': regex(CSS_COLOR_REGEX),
      '[color_options]': nullable(_validate_color_options),
      '[font_options]': nullable(_validate_font_options)
    }
  )
  if isinstance(spec.get('color_options'), list) and (spec['color'] not in spec['color_options']):
    raise Exception("default 'color' must be included in 'color_options' if provided")
  if isinstance(spec.get('font_options'), list) and (spec['font'] not in spec['font_options']):
    raise Exception("default 'font' must be included in 'font_options' if provided")

def _validate_video_spec(spec):
  validate(
    spec,
    {
      'resize': one_of(None, 'crop', 'scale'),
      'duration': { 'min_s': numeric_range('[0,∞)'), 'max_s': numeric_range('[1,∞)') },
      '[exclude_audio]': bool
    }
  )
  if spec['duration']['min_s'] > spec['duration']['max_s']:
    raise Exception("'min_s' cannot be greater than 'max_s' for an VIDEO input")

SPEC_FORMATS = {
  'AUDIO_CHOICE': _validate_audio_choice_spec,
  'IMG': dict,
  'IMG_GROUP': _validate_img_group_spec,
  'TEXT': _validate_text_spec,
  'VIDEO': _validate_video_spec
}

SECTION_FORMAT = {
  'section_id': non_empty_str,
  'owner_only': bool,
  'required': bool,
  'type': one_of(*SPEC_FORMATS.keys()),
  'spec': dict
}

def _validate_sections(sections):
  if not isinstance(sections, list): 
    raise Exception("'sections' should be a list")
  if not tuple(x for x in sections if not x['owner_only']):
    raise Exception("'sections' should contain at least one group input")

  section_ids = set()
  for i, section in enumerate(sections):
    section_id = section.get('section_id')
    if not section_id:
      raise Exception('section %i is missing an ID' % (i + 1))
    if section_id in section_ids:
      raise Exception("section IDs must be unique - duplicate on '%s'" % section_id)
    try:
      validate(section, SECTION_FORMAT)
      validate(section['spec'], SPEC_FORMATS[section['type']])
    except Exception as e:
      raise Exception("section '%s' is invalid" % section_id) from e


TEMPLATE_KEYS = {
  'name': non_empty_str,
  'description': non_empty_str,
  'sections': _validate_sections,
  'build': _validate_build_fn
}

def validate_template(template):
  for key, key_format in TEMPLATE_KEYS.items():
    value = getattr(template, key)
    validate(value, key_format)
